// -*- C++ -*-
#include "Rivet/Analysis.hh"
#include "Rivet/Projections/VetoedFinalState.hh"
#include "Rivet/Projections/IdentifiedFinalState.hh"
#include "Rivet/Projections/FromElectroweakDecay.hh"
#include "Rivet/Projections/DressedLeptons.hh"
#include "Rivet/Projections/FastJets.hh"
#include <iostream>
#include <map>
#include <cmath>
#include <bitset>


using namespace std;

namespace Rivet {

  /// @author W. H. Bell <W.Bell@cern.ch>
  /// @author A. Grohsjean <alexander.grohsjean@desy.de>
  class ATLAS_ttjets : public Analysis {
  public:
    ATLAS_ttjets(): 
      Analysis("ATLAS_ttjets"),
      m_jet_n(0),
      m_jet_ntag(0),
      m_met_et(0.),
      m_met_phi(0.),
      m_hMap(),
      m_chanLimit(3),
      m_histLimit(6)
    {}

  public:

    void init();
    void analyze(const Event& event);
    void finalize();
    
  private:

    // Event selection functions
    bool ejets(unsigned int &cutBits);
    bool mujets(unsigned int &cutBits);
    bool additionalJetCuts();

    // Histogram helper functions
    unsigned int thresholdLimit(unsigned int histId);
    unsigned int hIndex(unsigned int histId, unsigned int thresholdId);
    Histo1DPtr histogram(unsigned int histId, unsigned int thresholdId);
    double transMass(double ptLep, double phiLep, double met, double phiMet);
    unsigned int countJets(unsigned int ithres);

    // Objects that are used by the event selection decisions
    vector<DressedLepton> m_dressedelectrons;
    vector<DressedLepton> m_vetodressedelectrons;
    vector<DressedLepton> m_dressedmuons;
    vector<DressedLepton> m_vetodressedmuons;
    const Particles *m_neutrinos;
    const Jets *m_jets;
    unsigned int m_jet_n;
    unsigned int m_jet_ntag;
    double m_met_et;
    double m_met_phi;
    bool m_overlap;
    
    std::map<unsigned int, Histo1DPtr> m_hMap;
    unsigned int m_chanLimit;
    unsigned int m_histLimit;  
  };
  
  //-----------------------------------------------------------  
  
  void ATLAS_ttjets::init() {
    
    // All final state particles 
    std::vector<std::pair<double, double> > eta_full;
    eta_full.push_back(make_pair(-5.0,5.0));    
    FinalState fs(eta_full, 1.0*MeV); 
    
    // Get photons to dress leptons
    IdentifiedFinalState photons(fs);
    photons.acceptIdPair(PID::PHOTON);
    
    // Projection to find the electrons
    IdentifiedFinalState el_id(fs);
    el_id.acceptIdPair(PID::ELECTRON);
    FromElectroweakDecay electrons(el_id);
    electrons.acceptTauDecays(true);
    addProjection(electrons, "electrons");
    std::vector<std::pair<double, double> > eta_lep;
    eta_lep.push_back(make_pair(-2.5,2.5));    
    DressedLeptons dressedelectrons(photons, electrons, 0.1, true, eta_lep, 25.0*GeV,true);
    addProjection(dressedelectrons, "dressedelectrons");
    DressedLeptons vetodressedelectrons(photons, electrons, 0.1, true, eta_lep, 15.0*GeV,true);
    addProjection(vetodressedelectrons, "vetodressedelectrons");
    DressedLeptons ewdressedelectrons(photons, electrons, 0.1, true, eta_full, 1.0*MeV,true);
    addProjection(ewdressedelectrons, "ewdressedelectrons");
    
    // Projection to find the muons
    IdentifiedFinalState mu_id(fs);
    mu_id.acceptIdPair(PID::MUON);
    FromElectroweakDecay muons(mu_id);
    muons.acceptTauDecays(true);
    addProjection(muons, "muons");
    std::vector<std::pair<double, double> > eta_muon;
    DressedLeptons dressedmuons(photons, muons, 0.1, true, eta_lep, 25.0*GeV,true);
    addProjection(dressedmuons, "dressedmuons");
    DressedLeptons vetodressedmuons(photons, muons, 0.1, true, eta_lep, 15.0*GeV,true);
    addProjection(vetodressedmuons, "vetodressedmuons");
    DressedLeptons ewdressedmuons(photons, muons, 0.1, true, eta_full, 1.0*MeV,true);
    addProjection(ewdressedmuons, "ewdressedmuons");

    // Projection to find neutrinos and produce MET
    IdentifiedFinalState nu_id;
    nu_id.acceptNeutrinos();
    FromElectroweakDecay neutrinos(nu_id);
    neutrinos.acceptTauDecays(true);
    addProjection(neutrinos, "neutrinos"); 

    // Jet clustering.
    VetoedFinalState vfs;
    vfs.addVetoOnThisFinalState(ewdressedelectrons);
    vfs.addVetoOnThisFinalState(ewdressedmuons);
    vfs.addVetoOnThisFinalState(neutrinos);
    FastJets jets(vfs,FastJets::ANTIKT, 0.4);
    jets.useInvisibles();
    addProjection(jets, "jets");
    
    // Book histograms
    for(unsigned int ihist = 0; ihist < m_histLimit ; ihist++) {
      unsigned int threshLimit = thresholdLimit(ihist);
      for(unsigned int ithres = 0; ithres < threshLimit; ithres++) {
	histogram(ihist,ithres); // Create all histograms
      }
    }
  }
  
  //-----------------------------------------------------------
  
  void ATLAS_ttjets::analyze(const Event& event) {
  
    // Get the selected objects, using the projections.
    const vector<DressedLepton>& dressedelectrons = applyProjection<DressedLeptons>(event, "dressedelectrons").dressedLeptons();
    m_dressedelectrons=dressedelectrons;
    std::sort(m_dressedelectrons.begin(),m_dressedelectrons.end(),DressedLeptons::pTsorted); 
    const vector<DressedLepton>& vetodressedelectrons = applyProjection<DressedLeptons>(event, "vetodressedelectrons").dressedLeptons();
    m_vetodressedelectrons=vetodressedelectrons;
    
    const vector<DressedLepton>& dressedmuons = applyProjection<DressedLeptons>(event, "dressedmuons").dressedLeptons();
    m_dressedmuons=dressedmuons;
    std::sort(m_dressedmuons.begin(),m_dressedmuons.end(),DressedLeptons::pTsorted); 
    const vector<DressedLepton>& vetodressedmuons = applyProjection<DressedLeptons>(event, "vetodressedmuons").dressedLeptons();
    m_vetodressedmuons=vetodressedmuons;
    
    const Particles &neutrinos = applyProjection<FromElectroweakDecay>(event, "neutrinos").particlesByPt();
    m_neutrinos = &neutrinos;
    
    const Jets &jets = applyProjection<FastJets>(event, "jets").jetsByPt(25.0, MAXDOUBLE, -2.5, 2.5); 
    m_jets = &jets;
    
    const double weight = event.weight();

    // Calculate the missing ET, using the m_neutrinos.
    FourMomentum momentum;
    foreach (const Particle &p, (*m_neutrinos)) {
      const FourMomentum& mom = p.momentum();
      momentum += mom;
    }
    m_met_et = momentum.pT();
    m_met_phi = momentum.phi();

    // Check overlap of jets/leptons. 
    unsigned int i,j;
    m_jet_n = jets.size();
    m_jet_ntag = 0;
    m_overlap=false;
    for(i=0;i<m_jet_n;i++) {
      const Jet &jet = jets[i];
      // if dR(el,jet) < 0.4 skip the event
      foreach(const DressedLepton &el, dressedelectrons) {
	if(deltaR(jet.momentum(), el.momentum()) < 0.4) m_overlap=true; 
      }
      // if dR(mu,jet) < 0.4 skip the event
      foreach (const DressedLepton &mu, dressedmuons) {
	if(deltaR(jet.momentum(), mu.momentum()) < 0.4) m_overlap=true; 
      }
      // if dR(jet,jet) < 0.5 skip the event
      for(j=0;j<m_jet_n;j++) {
	const Jet &jet2 = jets[j];
	if(i == j) continue; // skip the diagonal
	if(deltaR(jet.momentum(), jet2.momentum()) < 0.5) m_overlap=true; 
      }
      // Count the number of b-tags
      if(jet.containsBottom()) {
	m_jet_ntag++;
      }
    }  
    
    // Evaluate basic event selection 
    unsigned int ejets_bits = 0, mujets_bits = 0;
    bool pass_ejets = ejets(ejets_bits);
    bool pass_mujets = mujets(mujets_bits);

    // Remove events with object overlap 
    if(m_overlap) vetoEvent;
    // basic event selection requirements
    if(!pass_ejets && !pass_mujets) vetoEvent; 

    // Check if the additional pT threshold requirements are passed
    bool pass_jetPt = additionalJetCuts();
    
    // Count the jet multiplicity for 25, 40, 60 and 80GeV
    unsigned int ithres, jet_n[4];
    for(ithres = 0; ithres < 4; ithres++) {
      jet_n[ithres] = countJets(ithres); 
    }
    
    // Fill histograms
    for(unsigned int ihist = 0; ihist < 6; ihist++) {
      if(ihist > 0 && !pass_jetPt) continue; // additional pT threshold cuts for pT plots
      unsigned int threshLimit = thresholdLimit(ihist);
      for(ithres = 0; ithres < threshLimit; ithres++) {
	if(jet_n[ithres] < 3) continue; // 3 or more jets for ljets 
	if(ihist == 0) {
	  histogram(ihist,ithres)->fill(jet_n[ithres],weight);  // njets
	}
	else if(ihist == 1) {
	  histogram(ihist,ithres)->fill(jets[0].pT(),weight); // leading jet pT 
	}
	else if(ihist == 2) {
	  histogram(ihist,ithres)->fill(jets[1].pT(),weight); // 2nd jet pT 
	}
	else if(ihist == 3 && jet_n[ithres] >= 3) {
	  histogram(ihist,ithres)->fill(jets[2].pT(),weight); // 3rd jet pT 
	}
	else if(ihist == 4 && jet_n[ithres] >= 4) {
	  histogram(ihist,ithres)->fill(jets[3].pT(),weight); // 4th jet pT 
	}
	else if(ihist == 5 && jet_n[ithres] >= 5) {
	  histogram(ihist,ithres)->fill(jets[4].pT(),weight); // 5th jet pT 
	}
      }
    } 
  }

  //-----------------------------------------------------------
  
  bool ATLAS_ttjets::ejets(unsigned int &cutBits) {  

    // 1. More than zero good electrons
    cutBits+=1; if(m_dressedelectrons.size() == 0) return false;
    
    // 2. No additional electrons passing the veto selection
    cutBits+=1<<1; if(m_vetodressedelectrons.size() > 1) return false;

    // 3. No muons passing the veto selection
    cutBits+=1<<2; if(m_vetodressedmuons.size() > 0) return false;
    
    // 4. total neutrino pT > 30GeV
    cutBits+=1<<3; if(m_met_et <= 30.0) return false;

    // 5. MTW > 35GeV 
    cutBits+=1<<4;
    if(transMass(m_dressedelectrons[0].momentum().pT(),
		 m_dressedelectrons[0].momentum().phi(),
		 m_met_et,
		 m_met_phi) <= 35.0) return false;
    

    // 6. At least one b-tagged jet
    cutBits+=1<<5; if(m_jet_ntag < 1) return false;

    // 7. At least three good jets
    cutBits+=1<<6; if(m_jet_n < 3) return false;

    cutBits+=1<<7;

    return true;
  }
  
  //-----------------------------------------------------------
  
  bool ATLAS_ttjets::mujets(unsigned int &cutBits){

    // 1. More than zero good muons
    cutBits+=1; if(m_dressedmuons.size() == 0) return false;

    // 2. No additional muons passing the veto selection
    cutBits+=1<<1; if(m_vetodressedmuons.size() > 1) return false;

    // 3. No electrons passing the veto selection
    cutBits+=1<<2; if(m_vetodressedelectrons.size() > 0) return false;

    // 4. total neutrino pT > 30GeV
    cutBits+=1<<3; if(m_met_et <= 30.0) return false;

    // 5. MTW > 35GeV 
    cutBits+=1<<4;
    if(transMass(m_dressedmuons[0].momentum().pT(),
		 m_dressedmuons[0].momentum().phi(),
		 m_met_et,
		 m_met_phi) <= 35.0) return false;
    

    // 6. At least one b-tagged jet
    cutBits+=1<<5; if(m_jet_ntag < 1) return false;

    // 7. At least three good jets
    cutBits+=1<<6; if(m_jet_n < 3) return false;

    cutBits+=1<<7;

    return true;
  }
  
  //-----------------------------------------------------------
  
  bool ATLAS_ttjets::additionalJetCuts(){
    
    if(m_jet_n < 2) return false;
    if(m_jets->at(0).pT() <= 50.0 || m_jets->at(1).pT() <= 35.0) return false;
    return true;
  }
  
  //-----------------------------------------------------------
  
  void ATLAS_ttjets::finalize() {
    
    double norm = crossSection()/sumOfWeights(); 
    std::map<unsigned int, Histo1DPtr>::iterator itr = m_hMap.begin();
    std::map<unsigned int, Histo1DPtr>::iterator itr_end = m_hMap.end();
    for(;itr!=itr_end;++itr) scale((*itr).second, norm);
    //average results  
    for(unsigned int ihist = 0; ihist < m_histLimit ; ihist++) {
      unsigned int threshLimit = thresholdLimit(ihist);
      for(unsigned int ithres = 0; ithres < threshLimit; ithres++){
	scale(histogram(ihist,ithres),0.5);
      }	
    }
  }
  
  //-----------------------------------------------------------
  
  unsigned int ATLAS_ttjets::thresholdLimit(unsigned int histId) {
  
    if(histId == 0) return 4;
    return 1;
  }
  
  //-----------------------------------------------------------
  
  unsigned int ATLAS_ttjets::hIndex(unsigned int histId, unsigned int thresholdId) {
    unsigned int hInd = 0;
    if(histId == 0) {
      hInd = thresholdId; 
    } else {
	hInd = thresholdLimit(0) + (histId-1) + thresholdId;
    }
    return hInd;
  }
  
  //-----------------------------------------------------------
  
  Histo1DPtr ATLAS_ttjets::histogram(unsigned int histId, unsigned int thresholdId) {
    
    assert(histId < m_histLimit);
    assert(thresholdId < thresholdLimit(histId));
    unsigned int hInd = hIndex(histId, thresholdId);
    
    if(m_hMap.find(hInd) != m_hMap.end()) return m_hMap[hInd];
    
    if(histId == 0) 
      m_hMap.insert(make_pair<unsigned int, Histo1DPtr>(hInd,bookHisto1D(1,thresholdId+1,1)));
    else
      m_hMap.insert(make_pair<unsigned int, Histo1DPtr>(hInd,bookHisto1D(2,histId,1)));
    
    return m_hMap[hInd];
  }
  
  //-----------------------------------------------------------
  
  double ATLAS_ttjets::transMass(double ptLep, double phiLep, double met, double phiMet) {
  
    return std::sqrt(2.0*ptLep*met*( 1 - std::cos( phiLep-phiMet ) ) );
  }
  
  //-----------------------------------------------------------
  
  unsigned int ATLAS_ttjets::countJets(unsigned int ithres) {
    
    if(ithres>4) assert(0);
    
    double pTcut[4] = {25.,40.,60.,80.};
    unsigned int i, jet_n = 0;
    for(i=0;i<m_jet_n;i++) {
      if(m_jets->at(i).pT() > pTcut[ithres]) jet_n++;
    }
    unsigned int ncutoff[4] = {8,7,6,5};
    if(jet_n>ncutoff[ithres]) jet_n=ncutoff[ithres]; 
    return jet_n;
  }
  
  //-----------------------------------------------------------
  
  // Declare the class as a hook for the plugin system
  DECLARE_RIVET_PLUGIN(ATLAS_ttjets);
}
